from __future__ import annotations

import uuid
from datetime import datetime
from typing import TYPE_CHECKING
from uuid import UUID

from dateutil.relativedelta import relativedelta
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import UniqueConstraint

from edc_document_status.model_mixins import DocumentStatusModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_metadata.model_mixins import MetadataHelperModelMixin
from edc_offstudy.model_mixins import OffstudyNonCrfModelMixin
from edc_timepoint.model_mixins import TimepointModelMixin
from edc_utils import formatted_datetime, to_utc
from edc_visit_schedule.model_mixins import VisitScheduleModelMixin
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.subject_schedule import NotOnScheduleError
from edc_visit_schedule.utils import is_baseline

from ..constants import CANCELLED_APPT, IN_PROGRESS_APPT
from ..exceptions import AppointmentDatetimeError, UnknownVisitCode
from ..managers import AppointmentManager
from ..utils import raise_on_appt_may_not_be_missed, update_appt_status
from .appointment_fields_model_mixin import AppointmentFieldsModelMixin
from .appointment_methods_model_mixin import AppointmentMethodsModelMixin
from .missed_appointment_model_mixin import MissedAppointmentModelMixin
from .window_period_model_mixin import WindowPeriodModelMixin

if TYPE_CHECKING:
    from edc_visit_schedule.schedule import Schedule

    from ..models import Appointment


class AppointmentModelMixin(
    NonUniqueSubjectIdentifierFieldMixin,
    AppointmentFieldsModelMixin,
    AppointmentMethodsModelMixin,
    TimepointModelMixin,
    MissedAppointmentModelMixin,
    WindowPeriodModelMixin,
    VisitScheduleModelMixin,
    DocumentStatusModelMixin,
    MetadataHelperModelMixin,
    OffstudyNonCrfModelMixin,
):
    """Mixin for the appointment model only.

    Only one appointment per subject visit+visit_code_sequence.

    Attribute 'visit_code_sequence' should be populated by the system.
    """

    metadata_helper_instance_attr = None

    offschedule_compare_dates_as_datetimes = False

    objects = AppointmentManager()

    def __str__(self) -> str:
        return f"{self.subject_identifier} {self.visit_code}.{self.visit_code_sequence}"

    def save(self: Appointment, *args, **kwargs):
        if not kwargs.get("update_fields"):
            if self.id and is_baseline(instance=self):
                visit_schedule = site_visit_schedules.get_visit_schedule(
                    self.visit_schedule_name
                )
                schedule: Schedule = visit_schedule.schedules.get(self.schedule_name)
                try:
                    onschedule_obj = django_apps.get_model(
                        schedule.onschedule_model
                    ).objects.get(
                        subject_identifier=self.subject_identifier,
                        onschedule_datetime__lte=to_utc(self.appt_datetime)
                        + relativedelta(seconds=1),
                    )
                except ObjectDoesNotExist as e:
                    dte_as_str = formatted_datetime(self.appt_datetime)
                    raise NotOnScheduleError(
                        "Subject is not on a schedule. Using subject_identifier="
                        f"`{self.subject_identifier}` and appt_datetime=`{dte_as_str}`."
                        f"Got {e}"
                    ) from e
                if self.appt_datetime > onschedule_obj.onschedule_datetime:
                    # update appointment timepoints
                    schedule.put_on_schedule(
                        subject_identifier=self.subject_identifier,
                        onschedule_datetime=to_utc(self.appt_datetime),
                        skip_baseline=True,
                    )
            else:
                # self.validate_appt_datetime_not_before_previous()
                self.validate_appt_datetime_not_after_next()
            raise_on_appt_may_not_be_missed(appointment=self)
            self.update_subject_visit_reason_or_raise()
            if self.appt_status != IN_PROGRESS_APPT and getattr(
                settings, "EDC_APPOINTMENT_CHECK_APPT_STATUS", True
            ):
                update_appt_status(self)
        super().save(*args, **kwargs)

    def natural_key(self) -> tuple:
        return (
            self.subject_identifier,
            self.visit_schedule_name,
            self.schedule_name,
            self.visit_code,
            self.visit_code_sequence,
        )

    @property
    def str_pk(self: Appointment) -> str | uuid.UUID:
        if isinstance(self.id, UUID):
            return str(self.pk)
        return self.pk

    def validate_appt_datetime_not_before_previous(self) -> None:
        if (
            self.appt_status != CANCELLED_APPT
            and self.appt_datetime
            and self.relative_previous
            and self.appt_datetime <= self.relative_previous.appt_datetime
        ):
            appt_datetime = formatted_datetime(self.appt_datetime)
            previous_appt_datetime = formatted_datetime(self.relative_previous.appt_datetime)
            raise AppointmentDatetimeError(
                "Datetime cannot be on or before previous appointment datetime. "
                f"Got {appt_datetime} <= {previous_appt_datetime}. "
                f"See appointment `{self}` and "
                f"`{self.relative_previous}`."
            )

    def validate_appt_datetime_not_after_next(self) -> None:
        if (
            self.appt_status != CANCELLED_APPT
            and self.appt_datetime
            and self.relative_next
            and self.appt_datetime >= self.relative_next.appt_datetime
        ):
            appt_datetime = formatted_datetime(self.appt_datetime)
            next_appt_datetime = formatted_datetime(self.relative_next.appt_datetime)
            raise AppointmentDatetimeError(
                "Datetime cannot be on or after next appointment datetime. "
                f"Got {appt_datetime} >= {next_appt_datetime}. "
                f"See appointment `{self}` and "
                f"`{self.relative_next}`."
            )

    @property
    def title(self: Appointment) -> str:
        if not self.schedule.visits.get(self.visit_code):
            valid_visit_codes = [v for v in self.schedule.visits]
            raise UnknownVisitCode(
                "Unknown visit code specified for existing apointment instance. "
                "Has the appointments schedule changed? Expected one of "
                f"{valid_visit_codes}. Got {self.visit_code}. "
                f"See {self}."
            )
        title = self.schedule.visits.get(self.visit_code).title
        if self.visit_code_sequence > 0:
            title = f"{title}.{self.visit_code_sequence}"
        return title

    @property
    def report_datetime(self: Appointment) -> datetime:
        return self.appt_datetime

    class Meta(NonUniqueSubjectIdentifierFieldMixin.Meta):
        abstract = True
        constraints = (
            UniqueConstraint(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "timepoint",
                    "visit_code_sequence",
                ],
                name="unique_%(app_label)s_%(class)s_100",
            ),
            UniqueConstraint(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "appt_datetime",
                ],
                name="unique_%(app_label)s_%(class)s_200",
            ),
        )
        indexes = (
            models.Index(fields=["appt_datetime"]),
            models.Index(fields=["appt_status"]),
            models.Index(fields=["timepoint", "visit_code_sequence"]),
            models.Index(fields=["subject_identifier", "appt_reason"]),
            models.Index(
                fields=[
                    "site",
                    "subject_identifier",
                    "timepoint",
                    "visit_code_sequence",
                ]
            ),
            models.Index(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "appt_reason",
                    "timepoint",
                    "visit_code_sequence",
                ]
            ),
        )



import os
from urllib.error import URLError
from urllib.request import urlretrieve

from django.conf import settings


def get_static_file(app_label: str, filename: str) -> str:
    path = os.path.join(settings.STATIC_ROOT or "", app_label, filename)
    if os.path.isfile(path):
        try:
            with open(path):
                pass
        except FileNotFoundError:
            path = os.path.join(f"https://{settings.STATIC_URL}", app_label, filename)
            try:
                urlretrieve(path)  # nosec B310
            except URLError:
                raise FileNotFoundError(
                    f"Static file not found. Tried "
                    f"STATIC_ROOT ({settings.STATIC_ROOT}) and "
                    f"STATIC_URL ({settings.STATIC_URL}). "
                    f"Got {app_label}/{filename}."
                )
    return path



from django.contrib.messages import get_messages


def message_in_queue(request, message_text):
    storage = get_messages(request)
    for message in storage:
        if message.message == message_text:
            return True
    return False



# Generated by Django 3.2.8 on 2022-02-15 21:08

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_auth", "0023_auto_20210423_1451"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userprofile",
            name="export_format",
            field=models.CharField(
                blank=True,
                choices=[
                    ("CSV", "CSV (delimited by pipe `|`)"),
                    (114, "Stata v10 or later"),
                    (117, "Stata v13 or later"),
                    (118, "Stata v14 or later"),
                    (119, "Stata v15 or later"),
                ],
                default="CSV",
                help_text="Note: requires export permissions",
                max_length=25,
                null=True,
                verbose_name="Export format",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-01-11 18:09
from __future__ import unicode_literals

from django.db import migrations, models

import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import django.utils.timezone

import edc_utils


class Migration(migrations.Migration):
    dependencies = [("edc_registration", "0004_auto_20161221_0018")]

    operations = [
        migrations.AlterField(
            model_name="registeredsubject",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="registeredsubject",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default="mac2-2.local",
                help_text="System field. (modified on create only)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="registeredsubject",
            name="hostname_modified",
            field=edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="registeredsubject",
            name="identity_or_pk",
            field=models.CharField(
                default=edc_utils.get_uuid,
                editable=False,
                max_length=75,
                unique=True,
                verbose_name="identity or pk",
            ),
        ),
        migrations.AlterField(
            model_name="registeredsubject",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="registeredsubject",
            name="relative_identifier",
            field=models.CharField(
                blank=True,
                help_text="For example, mother's identifier, if available / appropriate",
                max_length=36,
                null=True,
                verbose_name="Identifier of immediate relation",
            ),
        ),
        migrations.AlterField(
            model_name="registeredsubject",
            name="user_created",
            field=edc_model_fields.fields.userfield.UserField(
                blank=True, max_length=50, verbose_name="user created"
            ),
        ),
        migrations.AlterField(
            model_name="registeredsubject",
            name="user_modified",
            field=edc_model_fields.fields.userfield.UserField(
                blank=True, max_length=50, verbose_name="user modified"
            ),
        ),
    ]
